package com.earldouglas.rubble

import scala.collection.JavaConversions._
import java.io.InputStream
import scala.util.matching.Regex
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse
import javax.servlet.http.HttpServlet
import scala.xml.NodeSeq

object `package` {
  implicit def richRequest(req: HttpServletRequest) = new RichRequest(req)
  implicit def stringBody(body: String) = StringBody(body)
  implicit def streamBody(body: Stream[Byte]) = StreamBody(body)
  implicit def inputStreamBody(body: InputStream) = InputStreamBody(body)
  implicit def xmlBody(body: NodeSeq) = XmlBody(body)
  implicit def richServlet(servlet: HttpServlet) = new RichServlet(servlet)
  implicit def richResponse(res: HttpServletResponse) = new RichResponse(res)
  implicit def singleHeader(header: Header) = Seq(header)
  implicit def utf8ByteStream(string: String): Stream[Byte] = string.getBytes("UTF-8").toStream
  implicit def mapToHeaders(m: Map[String, String]): List[Header] = (m map { kv => Header(kv._1, kv._2) }).toList
  type Service = Function1[Request, Response]
}

trait Request {
  def method: String
  def root: String
  def uri: String
  def headers: Map[String, String]
  def params: Map[String, String]
  def body: Stream[Byte]
}

case class Response(status: Int = 200, headers: Map[String, String] = Map.empty, body: Stream[Byte] = Stream.empty)

class RichRequest(req: HttpServletRequest) extends Request {

  lazy val root = req.getScheme + "://" + req.getServerName + ":" + req.getServerPort + context

  lazy val context = Option(req.getContextPath).getOrElse("") +
                     Option(req.getServletPath).getOrElse("")

  lazy val headers =
    req.getHeaderNames.map(name => (name -> req.getHeader(name))).toMap

  lazy val method: String = req.getMethod

  lazy val uri: String = req.getRequestURI match {
    case x if x startsWith context => x.substring(context.length)
    case x => x
  }

  lazy val params: Map[String, String] =
    req.getParameterNames.map(name => (name -> req.getParameter(name))).toMap

  lazy val body: Stream[Byte] =
    Stream.continually(req.getInputStream.read).takeWhile(_ != -1).map(_.toByte)
}

class RichServlet(servlet: HttpServlet) {
  lazy val initParams: Map[String, String] =
    servlet.getInitParameterNames.map(name => (name -> servlet.getInitParameter(name))).toMap
}

class Purevlet extends HttpServlet {
  override def service(hreq: HttpServletRequest, hres: HttpServletResponse) {
    Option(getInitParameter("service")) match {
      case Some(svcClass) =>
        val svc = Class.forName(svcClass).newInstance.asInstanceOf[Service]
        val res = svc(hreq)
        hres.setStatus(res.status)
        res.headers foreach { kv => hres.setHeader(kv._1, kv._2) }
        val os = hres.getOutputStream
        res.body foreach { b => os.write(b) }
      case None =>
        hres.sendError(500, "no service specified")
    }
  }
}

case class Header(name: String, value: String)

sealed trait Body
case object NoBody extends Body
case class StringBody(body: String) extends Body
case class StreamBody(body: Stream[Byte]) extends Body
case class InputStreamBody(body: InputStream) extends Body
case class XmlBody(body: NodeSeq) extends Body
case class MustacheBody(resourceName: String, model: AnyRef) extends Body

class RichResponse(res: HttpServletResponse) {

  private implicit def moreFun[A](a: => A) =
    new {
      def ~>[B](f: A => B): B = f(a)
      def    <+(f: A => Any): A = { f(a); a }
    }

  def respond(r: Response) {
    respond(r.status, r.headers, r.body)
  }

  def respond(status: Int = 200, headers: Seq[Header] = Nil, body: Body = NoBody) {

    res.setStatus(status)

    headers.foreach(x => res.setHeader(x.name, x.value))

    if (Option(res.getContentType) == None)
      body match {
        case StringBody(body)        => res.setContentType("text/plain")
        case StreamBody(body)        => res.setContentType("application/octet+stream")
        case InputStreamBody(body)   => res.setContentType("application/octet+stream")
        case XmlBody(body)           => res.setContentType("text/html")
        case MustacheBody(name, mdl) => res.setContentType("text/html")
        case NoBody                  =>
      }

    body match {
      case StringBody(body)        => res.getWriter.write(body)
      case StreamBody(body)        => res.getOutputStream ~> { os => body foreach { b => os.write(b) } }
      case InputStreamBody(body)   => res.getOutputStream ~> { os => Iterator.continually(body.read).takeWhile(-1 !=).foreach(os.write) }
      case XmlBody(body)           => res.getWriter ~> { w => body.foreach(n => w.write(n.toString)) }
      case MustacheBody(name, mdl) => getClass.getClassLoader.getResource(name).getFile ~>
                                        { file => mustache.mustacheFactory.compile(file) } ~>
                                        { m => m.execute(res.getWriter, mdl).close }
      case NoBody                  =>
    }
  }

  def redirect(loc: String) = res <+ { _.setStatus(303) } <+ { _.setHeader("Location", loc) }
}
